@@ -3,7 +3,7 @@ use strict;
 use warnings;
 use Symbol;
 
-our $VERSION = '3.43_02';
+our $VERSION = '3.04';
 
 =head1 NAME
 
@@ -39,6 +39,4 @@ our @XSKeywords      = qw(
   EXPORT_XSUB_SYMBOLS
 );
 
-our $XSKeywordsAlternation = join('|', @XSKeywords);
-
 1;
