@@ -1,14 +1,13 @@
 package ExtUtils::ParseXS::CountLines;
 use strict;
 
-our $VERSION = '3.43_02';
+our $VERSION = '3.04';
 
 our $SECTION_END_MARKER;
 
 sub TIEHANDLE {
   my ($class, $cfile, $fh) = @_;
   $cfile =~ s/\\/\\\\/g;
-  $cfile =~ s/"/\\"/g;
   $SECTION_END_MARKER = qq{#line --- "$cfile"};
 
   return bless {
