@@ -3,18 +3,21 @@ use strict;
 use warnings;
 use Exporter;
 use File::Spec;
+use lib qw( lib );
 use ExtUtils::ParseXS::Constants ();
 
-our $VERSION = '3.43_02';
+our $VERSION = '3.04';
 
 our (@ISA, @EXPORT_OK);
 @ISA = qw(Exporter);
 @EXPORT_OK = qw(
   standard_typemap_locations
   trim_whitespace
+  tidy_type
   C_string
   valid_proto_string
   process_typemaps
+  make_targetable
   map_type
   standard_XS_defs
   assign_func_args
@@ -25,8 +28,6 @@ our (@ISA, @EXPORT_OK);
   blurt
   death
   check_conditional_preprocessor_statements
-  escape_file_for_line_directive
-  report_typemap_failure
 );
 
 =head1 NAME
@@ -38,9 +39,11 @@ ExtUtils::ParseXS::Utilities - Subroutines used with ExtUtils::ParseXS
   use ExtUtils::ParseXS::Utilities qw(
     standard_typemap_locations
     trim_whitespace
+    tidy_type
     C_string
     valid_proto_string
     process_typemaps
+    make_targetable
     map_type
     standard_XS_defs
     assign_func_args
@@ -50,8 +53,6 @@ ExtUtils::ParseXS::Utilities - Subroutines used with ExtUtils::ParseXS
     blurt
     death
     check_conditional_preprocessor_statements
-    escape_file_for_line_directive
-    report_typemap_failure
   );
 
 =head1 SUBROUTINES
@@ -124,35 +125,26 @@ Array holding list of directories to be searched for F<typemap> files.
 
 =cut
 
-SCOPE: {
-  my @tm_template;
+sub standard_typemap_locations {
+  my $include_ref = shift;
+  my @tm = qw(typemap);
 
-  sub standard_typemap_locations {
-    my $include_ref = shift;
-
-    if (not @tm_template) {
-      @tm_template = qw(typemap);
-
-      my $updir = File::Spec->updir();
-      foreach my $dir (
-          File::Spec->catdir(($updir) x 1),
-          File::Spec->catdir(($updir) x 2),
-          File::Spec->catdir(($updir) x 3),
-          File::Spec->catdir(($updir) x 4),
-      ) {
-        unshift @tm_template, File::Spec->catfile($dir, 'typemap');
-        unshift @tm_template, File::Spec->catfile($dir, lib => ExtUtils => 'typemap');
-      }
-    }
-
-    my @tm = @tm_template;
-    foreach my $dir (@{ $include_ref}) {
-      my $file = File::Spec->catfile($dir, ExtUtils => 'typemap');
-      unshift @tm, $file if -e $file;
-    }
-    return @tm;
+  my $updir = File::Spec->updir();
+  foreach my $dir (
+      File::Spec->catdir(($updir) x 1),
+      File::Spec->catdir(($updir) x 2),
+      File::Spec->catdir(($updir) x 3),
+      File::Spec->catdir(($updir) x 4),
+  ) {
+    unshift @tm, File::Spec->catfile($dir, 'typemap');
+    unshift @tm, File::Spec->catfile($dir, lib => ExtUtils => 'typemap');
   }
-} # end SCOPE
+  foreach my $dir (@{ $include_ref}) {
+    my $file = File::Spec->catfile($dir, ExtUtils => 'typemap');
+    unshift @tm, $file if -e $file;
+  }
+  return @tm;
+}
 
 =head2 C<trim_whitespace()>
 
@@ -179,6 +171,45 @@ sub trim_whitespace {
   $_[0] =~ s/^\s+|\s+$//go;
 }
 
+=head2 C<tidy_type()>
+
+=over 4
+
+=item * Purpose
+
+Rationalize any asterisks (C<*>) by joining them into bunches, removing
+interior whitespace, then trimming leading and trailing whitespace.
+
+=item * Arguments
+
+    ($ret_type) = tidy_type($_);
+
+String to be cleaned up.
+
+=item * Return Value
+
+String cleaned up.
+
+=back
+
+=cut
+
+sub tidy_type {
+  local ($_) = @_;
+
+  # rationalise any '*' by joining them into bunches and removing whitespace
+  s#\s*(\*+)\s*#$1#g;
+  s#(\*+)# $1 #g;
+
+  # change multiple whitespace into a single space
+  s/\s+/ /g;
+
+  # trim leading & trailing whitespace
+  trim_whitespace($_);
+
+  $_;
+}
+
 =head2 C<C_string()>
 
 =over 4
@@ -231,7 +262,7 @@ Upon failure, returns C<0>.
 =cut
 
 sub valid_proto_string {
-  my ($string) = @_;
+  my($string) = @_;
 
   if ( $string =~ /^$ExtUtils::ParseXS::Constants::PrototypeRegexp+$/ ) {
     return $string;
@@ -251,7 +282,7 @@ Process all typemap files.
 =item * Arguments
 
   my $typemaps_object = process_typemaps( $args{typemap}, $pwd );
-
+      
 List of two elements:  C<typemap> element from C<%args>; current working
 directory.
 
@@ -288,6 +319,76 @@ sub process_typemaps {
   return $typemap;
 }
 
+=head2 C<make_targetable()>
+
+=over 4
+
+=item * Purpose
+
+Populate C<%targetable>.  This constitutes a refinement of the output of
+C<process_typemaps()> with respect to its fourth output, C<$output_expr_ref>.
+
+=item * Arguments
+
+  %targetable = make_targetable($output_expr_ref);
+      
+Single hash reference:  the fourth such ref returned by C<process_typemaps()>.
+
+=item * Return Value
+
+Hash.
+
+=back
+
+=cut
+
+sub make_targetable {
+  my $output_expr_ref = shift;
+
+  our $bal; # ()-balanced
+  $bal = qr[
+    (?:
+      (?>[^()]+)
+      |
+      \( (??{ $bal }) \)
+    )*
+  ]x;
+
+  # matches variations on (SV*)
+  my $sv_cast = qr[
+    (?:
+      \( \s* SV \s* \* \s* \) \s*
+    )?
+  ]x;
+
+  my $size = qr[ # Third arg (to setpvn)
+    , \s* (??{ $bal })
+  ]x;
+
+  my %targetable;
+  foreach my $key (keys %{ $output_expr_ref }) {
+    # We can still bootstrap compile 're', because in code re.pm is
+    # available to miniperl, and does not attempt to load the XS code.
+    use re 'eval';
+
+    my ($type, $with_size, $arg, $sarg) =
+      ($output_expr_ref->{$key} =~
+        m[^
+          \s+
+          sv_set([iunp])v(n)?    # Type, is_setpvn
+          \s*
+          \( \s*
+            $sv_cast \$arg \s* , \s*
+            ( (??{ $bal }) )    # Set from
+          ( (??{ $size }) )?    # Possible sizeof set-from
+          \) \s* ; \s* $
+        ]x
+    );
+    $targetable{$key} = [$type, $with_size, $arg, $sarg] if $type;
+  }
+  return %targetable;
+}
+
 =head2 C<map_type()>
 
 =over 4
@@ -314,7 +415,7 @@ sub map_type {
   my ($self, $type, $varname) = @_;
 
   # C++ has :: in types too so skip this
-  $type =~ tr/:/_/ unless $self->{RetainCplusplusHierarchicalTypes};
+  $type =~ tr/:/_/ unless $self->{hiertype};
   $type =~ s/^array\(([^,]*),(.*)\).*/$1 */s;
   if ($varname) {
     if ($type =~ / \( \s* \* (?= \s* \) ) /xg) {
@@ -354,96 +455,14 @@ sub standard_XS_defs {
 #  define PERL_UNUSED_VAR(var) if (0) var = var
 #endif
 
-#ifndef dVAR
-#  define dVAR		dNOOP
-#endif
-
-
-/* This stuff is not part of the API! You have been warned. */
-#ifndef PERL_VERSION_DECIMAL
-#  define PERL_VERSION_DECIMAL(r,v,s) (r*1000000 + v*1000 + s)
-#endif
-#ifndef PERL_DECIMAL_VERSION
-#  define PERL_DECIMAL_VERSION \\
-	  PERL_VERSION_DECIMAL(PERL_REVISION,PERL_VERSION,PERL_SUBVERSION)
-#endif
-#ifndef PERL_VERSION_GE
-#  define PERL_VERSION_GE(r,v,s) \\
-	  (PERL_DECIMAL_VERSION >= PERL_VERSION_DECIMAL(r,v,s))
-#endif
-#ifndef PERL_VERSION_LE
-#  define PERL_VERSION_LE(r,v,s) \\
-	  (PERL_DECIMAL_VERSION <= PERL_VERSION_DECIMAL(r,v,s))
-#endif
-
-/* XS_INTERNAL is the explicit static-linkage variant of the default
- * XS macro.
- *
- * XS_EXTERNAL is the same as XS_INTERNAL except it does not include
- * "STATIC", ie. it exports XSUB symbols. You probably don't want that
- * for anything but the BOOT XSUB.
- *
- * See XSUB.h in core!
- */
-
-
-/* TODO: This might be compatible further back than 5.10.0. */
-#if PERL_VERSION_GE(5, 10, 0) && PERL_VERSION_LE(5, 15, 1)
-#  undef XS_EXTERNAL
-#  undef XS_INTERNAL
-#  if defined(__CYGWIN__) && defined(USE_DYNAMIC_LOADING)
-#    define XS_EXTERNAL(name) __declspec(dllexport) XSPROTO(name)
-#    define XS_INTERNAL(name) STATIC XSPROTO(name)
-#  endif
-#  if defined(__SYMBIAN32__)
-#    define XS_EXTERNAL(name) EXPORT_C XSPROTO(name)
-#    define XS_INTERNAL(name) EXPORT_C STATIC XSPROTO(name)
-#  endif
-#  ifndef XS_EXTERNAL
-#    if defined(HASATTRIBUTE_UNUSED) && !defined(__cplusplus)
-#      define XS_EXTERNAL(name) void name(pTHX_ CV* cv __attribute__unused__)
-#      define XS_INTERNAL(name) STATIC void name(pTHX_ CV* cv __attribute__unused__)
-#    else
-#      ifdef __cplusplus
-#        define XS_EXTERNAL(name) extern "C" XSPROTO(name)
-#        define XS_INTERNAL(name) static XSPROTO(name)
-#      else
-#        define XS_EXTERNAL(name) XSPROTO(name)
-#        define XS_INTERNAL(name) STATIC XSPROTO(name)
-#      endif
-#    endif
-#  endif
-#endif
-
-/* perl >= 5.10.0 && perl <= 5.15.1 */
-
-
-/* The XS_EXTERNAL macro is used for functions that must not be static
+/* Starting from 5.15.2, XS(name) defines a static function (==internal)
+ * and the XS_EXTERNAL macro is used for functions that must not be static
  * like the boot XSUB of a module. If perl didn't have an XS_EXTERNAL
  * macro defined, the best we can do is assume XS is the same.
- * Dito for XS_INTERNAL.
  */
 #ifndef XS_EXTERNAL
 #  define XS_EXTERNAL(name) XS(name)
 #endif
-#ifndef XS_INTERNAL
-#  define XS_INTERNAL(name) XS(name)
-#endif
-
-/* Now, finally, after all this mess, we want an ExtUtils::ParseXS
- * internal macro that we're free to redefine for varying linkage due
- * to the EXPORT_XSUB_SYMBOLS XS keyword. This is internal, use
- * XS_EXTERNAL(name) or XS_INTERNAL(name) in your code if you need to!
- */
-
-#undef XS_EUPXS
-#if defined(PERL_EUPXS_ALWAYS_EXPORT)
-#  define XS_EUPXS(name) XS_EXTERNAL(name)
-#else
-   /* default to internal */
-#  define XS_EUPXS(name) XS_INTERNAL(name)
-#endif
-
 EOF
 
   print <<"EOF";
@@ -452,10 +471,10 @@ EOF
 
 /* prototype to pass -Wmissing-prototypes */
 STATIC void
-S_croak_xs_usage(const CV *const cv, const char *const params);
+S_croak_xs_usage(pTHX_ const CV *const cv, const char *const params);
 
 STATIC void
-S_croak_xs_usage(const CV *const cv, const char *const params)
+S_croak_xs_usage(pTHX_ const CV *const cv, const char *const params)
 {
     const GV *const gv = CvGV(cv);
 
@@ -467,17 +486,21 @@ S_croak_xs_usage(const CV *const cv, const char *const params)
         const char *const hvname = stash ? HvNAME(stash) : NULL;
 
         if (hvname)
-	    Perl_croak_nocontext("Usage: %s::%s(%s)", hvname, gvname, params);
+            Perl_croak(aTHX_ "Usage: %s::%s(%s)", hvname, gvname, params);
         else
-	    Perl_croak_nocontext("Usage: %s(%s)", gvname, params);
+            Perl_croak(aTHX_ "Usage: %s(%s)", gvname, params);
     } else {
         /* Pants. I don't think that it should be possible to get here. */
-	Perl_croak_nocontext("Usage: CODE(0x%" UVxf ")(%s)", PTR2UV(cv), params);
+        Perl_croak(aTHX_ "Usage: CODE(0x%"UVxf")(%s)", PTR2UV(cv), params);
     }
 }
 #undef  PERL_ARGS_ASSERT_CROAK_XS_USAGE
 
+#ifdef PERL_IMPLICIT_CONTEXT
+#define croak_xs_usage(a,b)    S_croak_xs_usage(aTHX_ a,b)
+#else
 #define croak_xs_usage        S_croak_xs_usage
+#endif
 
 #endif
 
@@ -490,12 +513,6 @@ S_croak_xs_usage(const CV *const cv, const char *const params)
 #define newXSproto_portable(name, c_impl, file, proto) (PL_Sv=(SV*)newXS(name, c_impl, file), sv_setpv(PL_Sv, proto), (CV*)PL_Sv)
 #endif /* !defined(newXS_flags) */
 
-#if PERL_VERSION_LE(5, 21, 5)
-#  define newXS_deffile(a,b) Perl_newXS(aTHX_ a,b,file)
-#else
-#  define newXS_deffile(a,b) Perl_newXS_deffile(aTHX_ a,b)
-#endif
-
 EOF
   return 1;
 }
@@ -568,7 +585,7 @@ sub analyze_preprocessor_statements {
     push(@{ $self->{XSStack} }, {type => 'if'});
   }
   else {
-    $self->death("Error: '$statement' with no matching 'if'")
+    $self->death("Error: `$statement' with no matching `if'")
       if $self->{XSStack}->[-1]{type} ne 'if';
     if ($self->{XSStack}->[-1]{varname}) {
       push(@{ $self->{InitFileCode} }, "#endif\n");
@@ -745,77 +762,6 @@ sub check_conditional_preprocessor_statements {
   }
 }
 
-=head2 C<escape_file_for_line_directive()>
-
-=over 4
-
-=item * Purpose
-
-Escapes a given code source name (typically a file name but can also
-be a command that was read from) so that double-quotes and backslashes are escaped.
-
-=item * Arguments
-
-A string.
-
-=item * Return Value
-
-A string with escapes for double-quotes and backslashes.
-
-=back
-
-=cut
-
-sub escape_file_for_line_directive {
-  my $string = shift;
-  $string =~ s/\\/\\\\/g;
-  $string =~ s/"/\\"/g;
-  return $string;
-}
-
-=head2 C<report_typemap_failure>
-
-=over 4
-
-=item * Purpose
-
-Do error reporting for missing typemaps.
-
-=item * Arguments
-
-The C<ExtUtils::ParseXS> object.
-
-An C<ExtUtils::Typemaps> object.
-
-The string that represents the C type that was not found in the typemap.
-
-Optionally, the string C<death> or C<blurt> to choose
-whether the error is immediately fatal or not. Default: C<blurt>
-
-=item * Return Value
-
-Returns nothing. Depending on the arguments, this
-may call C<death> or C<blurt>, the former of which is
-fatal.
-
-=back
-
-=cut
-
-sub report_typemap_failure {
-  my ($self, $tm, $ctype, $error_method) = @_;
-  $error_method ||= 'blurt';
-
-  my @avail_ctypes = $tm->list_mapped_ctypes;
-
-  my $err = "Could not find a typemap for C type '$ctype'.\n"
-            . "The following C types are mapped by the current typemap:\n'"
-            . join("', '", @avail_ctypes) . "'\n";
-
-  $self->$error_method($err);
-  return();
-}
-
 1;
 
 # vim: ts=2 sw=2 et:
